@@ -18,7 +18,7 @@
 %                            R  R    GGG   BBBB                               %
 %                                                                             %
 %                                                                             %
-%                   Read/Write GraphicsMagick Image Format.                   %
+%                    Read/Write Raw RGB Image Format.                         %
 %                                                                             %
 %                                                                             %
 %                              Software Design                                %
@@ -37,7 +37,7 @@
 */
 #include "magick/studio.h"
 #include "magick/blob.h"
-#include "magick/cache.h"
+#include "magick/pixel_cache.h"
 #include "magick/constitute.h"
 #include "magick/magick.h"
 #include "magick/monitor.h"
@@ -104,8 +104,9 @@ static Image *ReadRGBImage(const ImageInfo *image_info,ExceptionInfo *exception)
   unsigned int
     status;
 
-  unsigned long
-    packet_size;
+  unsigned int
+    packet_size,
+    quantum_size;
 
   assert(image_info != (const ImageInfo *) NULL);
   assert(image_info->signature == MagickSignature);
@@ -113,7 +114,7 @@ static Image *ReadRGBImage(const ImageInfo *image_info,ExceptionInfo *exception)
   assert(exception->signature == MagickSignature);
   image=AllocateImage(image_info);
   if ((image->columns == 0) || (image->rows == 0))
-    ThrowReaderException(OptionError,"MustSpecifyImageSize",image);
+    ThrowReaderException(OptionError,MustSpecifyImageSize,image);
   if (image_info->interlace != PartitionInterlace)
     {
       /*
@@ -121,23 +122,34 @@ static Image *ReadRGBImage(const ImageInfo *image_info,ExceptionInfo *exception)
       */
       status=OpenBlob(image_info,image,ReadBinaryBlobMode,exception);
       if (status == False)
-        ThrowReaderException(FileOpenError,"UnableToOpenFile",image);
+        ThrowReaderException(FileOpenError,UnableToOpenFile,image);
       for (i=0; i < image->offset; i++)
-        (void) ReadBlobByte(image);
+        {
+          if (EOF == ReadBlobByte(image))
+            ThrowException(exception,CorruptImageError,UnexpectedEndOfFile,
+                           image->filename);
+        }
     }
   /*
     Allocate memory for a scanline.
   */
-  packet_size=image->depth > 8 ? 6 : 3;
+  if (image->depth <= 8)
+    quantum_size=8;
+  else if (image->depth <= 16)
+    quantum_size=16;
+  else
+    quantum_size=32;
+
+  packet_size=(quantum_size*3)/8;
   if (LocaleCompare(image_info->magick,"RGBA") == 0)
     {
       image->matte=True;
-      packet_size=image->depth > 8 ? 8 : 4;
+      packet_size=(quantum_size*4)/8;
     }
-  scanline=(unsigned char *)
-    AcquireMemory(packet_size*image->tile_info.width);
+  scanline=MagickAllocateMemory(unsigned char *,
+    packet_size*image->tile_info.width);
   if (scanline == (unsigned char *) NULL)
-    ThrowReaderException(ResourceLimitError,"MemoryAllocationFailed",image);
+    ThrowReaderException(ResourceLimitError,MemoryAllocationFailed,image);
   if (image_info->subrange != 0)
     while (image->scene < image_info->subimage)
     {
@@ -175,9 +187,9 @@ static Image *ReadRGBImage(const ImageInfo *image_info,ExceptionInfo *exception)
           if (q == (PixelPacket *) NULL)
             break;
           if (!image->matte)
-            (void) PushImagePixels(image,RGBQuantum,scanline+x);
+            (void) ImportImagePixelArea(image,RGBQuantum,quantum_size,scanline+x,0,0);
           else
-            (void) PushImagePixels(image,RGBAQuantum,scanline+x);
+            (void) ImportImagePixelArea(image,RGBAQuantum,quantum_size,scanline+x,0,0);
           if (!SyncImagePixels(image))
             break;
           if (image->previous == (Image *) NULL)
@@ -195,7 +207,7 @@ static Image *ReadRGBImage(const ImageInfo *image_info,ExceptionInfo *exception)
         /*
           Line interlacing:  RRR...GGG...BBB...RRR...GGG...BBB...
         */
-        packet_size=image->depth > 8 ? 2 : 1;
+        packet_size=(quantum_size)/8;
         for (y=0; y < image->tile_info.y; y++)
           (void) ReadBlob(image,packet_size*image->tile_info.width,scanline);
         for (y=0; y < (long) image->rows; y++)
@@ -205,16 +217,16 @@ static Image *ReadRGBImage(const ImageInfo *image_info,ExceptionInfo *exception)
           q=SetImagePixels(image,0,y,image->columns,1);
           if (q == (PixelPacket *) NULL)
             break;
-          (void) PushImagePixels(image,RedQuantum,scanline+x);
+          (void) ImportImagePixelArea(image,RedQuantum,quantum_size,scanline+x,0,0);
           (void) ReadBlob(image,packet_size*image->tile_info.width,scanline);
-          (void) PushImagePixels(image,GreenQuantum,scanline+x);
+          (void) ImportImagePixelArea(image,GreenQuantum,quantum_size,scanline+x,0,0);
           (void) ReadBlob(image,packet_size*image->tile_info.width,scanline);
-          (void) PushImagePixels(image,BlueQuantum,scanline+x);
+          (void) ImportImagePixelArea(image,BlueQuantum,quantum_size,scanline+x,0,0);
           if (image->matte)
             {
               (void) ReadBlob(image,packet_size*image->tile_info.width,
                 scanline);
-              (void) PushImagePixels(image,AlphaQuantum,scanline+x);
+              (void) ImportImagePixelArea(image,AlphaQuantum,quantum_size,scanline+x,0,0);
             }
           if (!SyncImagePixels(image))
             break;
@@ -242,9 +254,9 @@ static Image *ReadRGBImage(const ImageInfo *image_info,ExceptionInfo *exception)
             AppendImageFormat("R",image->filename);
             status=OpenBlob(image_info,image,ReadBinaryBlobMode,exception);
             if (status == False)
-              ThrowReaderException(FileOpenError,"UnableToOpenFile",image);
+              ThrowReaderException(FileOpenError,UnableToOpenFile,image);
           }
-        packet_size=image->depth > 8 ? 2 : 1;
+        packet_size=(quantum_size)/8;
         for (y=0; y < image->tile_info.y; y++)
           (void) ReadBlob(image,packet_size*image->tile_info.width,scanline);
         i=0;
@@ -256,7 +268,7 @@ static Image *ReadRGBImage(const ImageInfo *image_info,ExceptionInfo *exception)
           q=SetImagePixels(image,0,y,image->columns,1);
           if (q == (PixelPacket *) NULL)
             break;
-          (void) PushImagePixels(image,RedQuantum,scanline+x);
+          (void) ImportImagePixelArea(image,RedQuantum,quantum_size,scanline+x,0,0);
           if (!SyncImagePixels(image))
             break;
           if (image->previous == (Image *) NULL)
@@ -274,7 +286,7 @@ static Image *ReadRGBImage(const ImageInfo *image_info,ExceptionInfo *exception)
             AppendImageFormat("G",image->filename);
             status=OpenBlob(image_info,image,ReadBinaryBlobMode,exception);
             if (status == False)
-              ThrowReaderException(FileOpenError,"UnableToOpenFile",image);
+              ThrowReaderException(FileOpenError,UnableToOpenFile,image);
           }
         for (y=0; y < image->tile_info.y; y++)
           (void) ReadBlob(image,packet_size*image->tile_info.width,scanline);
@@ -284,7 +296,7 @@ static Image *ReadRGBImage(const ImageInfo *image_info,ExceptionInfo *exception)
           q=GetImagePixels(image,0,y,image->columns,1);
           if (q == (PixelPacket *) NULL)
             break;
-          (void) PushImagePixels(image,GreenQuantum,scanline+x);
+          (void) ImportImagePixelArea(image,GreenQuantum,quantum_size,scanline+x,0,0);
           if (!SyncImagePixels(image))
             break;
           if (image->previous == (Image *) NULL)
@@ -302,7 +314,7 @@ static Image *ReadRGBImage(const ImageInfo *image_info,ExceptionInfo *exception)
             AppendImageFormat("B",image->filename);
             status=OpenBlob(image_info,image,ReadBinaryBlobMode,exception);
             if (status == False)
-              ThrowReaderException(FileOpenError,"UnableToOpenFile",image);
+              ThrowReaderException(FileOpenError,UnableToOpenFile,image);
           }
         for (y=0; y < image->tile_info.y; y++)
           (void) ReadBlob(image,packet_size*image->tile_info.width,scanline);
@@ -312,7 +324,7 @@ static Image *ReadRGBImage(const ImageInfo *image_info,ExceptionInfo *exception)
           q=GetImagePixels(image,0,y,image->columns,1);
           if (q == (PixelPacket *) NULL)
             break;
-          (void) PushImagePixels(image,BlueQuantum,scanline+x);
+          (void) ImportImagePixelArea(image,BlueQuantum,quantum_size,scanline+x,0,0);
           if (!SyncImagePixels(image))
             break;
           if (image->previous == (Image *) NULL)
@@ -335,7 +347,7 @@ static Image *ReadRGBImage(const ImageInfo *image_info,ExceptionInfo *exception)
                 AppendImageFormat("A",image->filename);
                 status=OpenBlob(image_info,image,ReadBinaryBlobMode,exception);
                 if (status == False)
-                  ThrowReaderException(FileOpenError,"UnableToOpenFile",image);
+                  ThrowReaderException(FileOpenError,UnableToOpenFile,image);
               }
             for (y=0; y < image->tile_info.y; y++)
               (void) ReadBlob(image,packet_size*image->tile_info.width,
@@ -347,7 +359,7 @@ static Image *ReadRGBImage(const ImageInfo *image_info,ExceptionInfo *exception)
               q=GetImagePixels(image,0,y,image->columns,1);
               if (q == (PixelPacket *) NULL)
                 break;
-              (void) PushImagePixels(image,AlphaQuantum,scanline+x);
+              (void) ImportImagePixelArea(image,AlphaQuantum,quantum_size,scanline+x,0,0);
               if (!SyncImagePixels(image))
                 break;
               if (image->previous == (Image *) NULL)
@@ -362,13 +374,13 @@ static Image *ReadRGBImage(const ImageInfo *image_info,ExceptionInfo *exception)
                 scanline);
           }
         if (image_info->interlace == PartitionInterlace)
-          (void) strncpy(image->filename,image_info->filename,MaxTextExtent-1);
+          (void) strlcpy(image->filename,image_info->filename,MaxTextExtent);
         break;
       }
     }
     if (EOFBlob(image))
       {
-        ThrowException(exception,CorruptImageError,"UnexpectedEndOfFile",
+        ThrowException(exception,CorruptImageError,UnexpectedEndOfFile,
           image->filename);
         break;
       }
@@ -397,7 +409,7 @@ static Image *ReadRGBImage(const ImageInfo *image_info,ExceptionInfo *exception)
           break;
       }
   } while (count != 0);
-  LiberateMemory((void **) &scanline);
+  MagickFreeMemory(scanline);
   while (image->previous != (Image *) NULL)
     image=image->previous;
   CloseBlob(image);
@@ -517,8 +529,9 @@ static unsigned int WriteRGBImage(const ImageInfo *image_info,Image *image)
   unsigned int
     status;
 
-  unsigned long
+  unsigned int
     packet_size,
+    quantum_size,
     scene;
 
   /*
@@ -528,12 +541,20 @@ static unsigned int WriteRGBImage(const ImageInfo *image_info,Image *image)
   assert(image_info->signature == MagickSignature);
   assert(image != (Image *) NULL);
   assert(image->signature == MagickSignature);
-  packet_size=image->depth > 8 ? 6 : 3;
+
+  if (image->depth <= 8)
+    quantum_size=8;
+  else if (image->depth <= 16)
+    quantum_size=16;
+  else
+    quantum_size=32;
+
+  packet_size=(quantum_size*3)/8;
   if (LocaleCompare(image_info->magick,"RGBA") == 0)
-    packet_size=image->depth > 8 ? 8 : 4;
-  pixels=(unsigned char *) AcquireMemory(packet_size*image->columns);
+    packet_size=(quantum_size*4)/8;
+  pixels=MagickAllocateMemory(unsigned char *,packet_size*image->columns);
   if (pixels == (unsigned char *) NULL)
-    ThrowWriterException(ResourceLimitError,"MemoryAllocationFailed",image);
+    ThrowWriterException(ResourceLimitError,MemoryAllocationFailed,image);
   if (image_info->interlace != PartitionInterlace)
     {
       /*
@@ -541,7 +562,7 @@ static unsigned int WriteRGBImage(const ImageInfo *image_info,Image *image)
       */
       status=OpenBlob(image_info,image,WriteBinaryBlobMode,&image->exception);
       if (status == False)
-        ThrowWriterException(FileOpenError,"UnableToOpenFile",image);
+        ThrowWriterException(FileOpenError,UnableToOpenFile,image);
     }
   scene=0;
   do
@@ -549,7 +570,7 @@ static unsigned int WriteRGBImage(const ImageInfo *image_info,Image *image)
     /*
       Convert MIFF to RGB raster pixels.
     */
-    TransformColorspace(image,RGBColorspace);
+    (void) TransformColorspace(image,RGBColorspace);
     if (LocaleCompare(image_info->magick,"RGBA") == 0)
       if (!image->matte)
         SetImageOpacity(image,OpaqueOpacity);
@@ -568,12 +589,12 @@ static unsigned int WriteRGBImage(const ImageInfo *image_info,Image *image)
             break;
           if (LocaleCompare(image_info->magick,"RGBA") != 0)
             {
-              (void) PopImagePixels(image,RGBQuantum,pixels);
+              (void) ExportImagePixelArea(image,RGBQuantum,quantum_size,pixels,0,0);
               (void) WriteBlob(image,packet_size*image->columns,pixels);
             }
           else
             {
-              (void) PopImagePixels(image,RGBAQuantum,pixels);
+              (void) ExportImagePixelArea(image,RGBAQuantum,quantum_size,pixels,0,0);
               (void) WriteBlob(image,packet_size*image->columns,pixels);
             }
           if (image->previous == (Image *) NULL)
@@ -593,15 +614,15 @@ static unsigned int WriteRGBImage(const ImageInfo *image_info,Image *image)
           p=AcquireImagePixels(image,0,y,image->columns,1,&image->exception);
           if (p == (const PixelPacket *) NULL)
             break;
-          (void) PopImagePixels(image,RedQuantum,pixels);
+          (void) ExportImagePixelArea(image,RedQuantum,quantum_size,pixels,0,0);
           (void) WriteBlob(image,image->columns,pixels);
-          (void) PopImagePixels(image,GreenQuantum,pixels);
+          (void) ExportImagePixelArea(image,GreenQuantum,quantum_size,pixels,0,0);
           (void) WriteBlob(image,image->columns,pixels);
-          (void) PopImagePixels(image,BlueQuantum,pixels);
+          (void) ExportImagePixelArea(image,BlueQuantum,quantum_size,pixels,0,0);
           (void) WriteBlob(image,image->columns,pixels);
           if (LocaleCompare(image_info->magick,"RGBA") == 0)
             {
-              (void) PopImagePixels(image,AlphaQuantum,pixels);
+              (void) ExportImagePixelArea(image,AlphaQuantum,quantum_size,pixels,0,0);
               (void) WriteBlob(image,image->columns,pixels);
             }
           if (QuantumTick(y,image->rows))
@@ -621,14 +642,14 @@ static unsigned int WriteRGBImage(const ImageInfo *image_info,Image *image)
             AppendImageFormat("R",image->filename);
             status=OpenBlob(image_info,image,WriteBinaryBlobMode,&image->exception);
             if (status == False)
-              ThrowWriterException(FileOpenError,"UnableToOpenFile",image);
+              ThrowWriterException(FileOpenError,UnableToOpenFile,image);
           }
         for (y=0; y < (long) image->rows; y++)
         {
           p=AcquireImagePixels(image,0,y,image->columns,1,&image->exception);
           if (p == (const PixelPacket *) NULL)
             break;
-          (void) PopImagePixels(image,RedQuantum,pixels);
+          (void) ExportImagePixelArea(image,RedQuantum,quantum_size,pixels,0,0);
           (void) WriteBlob(image,image->columns,pixels);
         }
         if (image_info->interlace == PartitionInterlace)
@@ -637,7 +658,7 @@ static unsigned int WriteRGBImage(const ImageInfo *image_info,Image *image)
             AppendImageFormat("G",image->filename);
             status=OpenBlob(image_info,image,WriteBinaryBlobMode,&image->exception);
             if (status == False)
-              ThrowWriterException(FileOpenError,"UnableToOpenFile",image);
+              ThrowWriterException(FileOpenError,UnableToOpenFile,image);
           }
         if (!MagickMonitor(SaveImageText,100,400,&image->exception))
           break;
@@ -646,7 +667,7 @@ static unsigned int WriteRGBImage(const ImageInfo *image_info,Image *image)
           p=AcquireImagePixels(image,0,y,image->columns,1,&image->exception);
           if (p == (const PixelPacket *) NULL)
             break;
-          (void) PopImagePixels(image,GreenQuantum,pixels);
+          (void) ExportImagePixelArea(image,GreenQuantum,quantum_size,pixels,0,0);
           (void) WriteBlob(image,image->columns,pixels);
         }
         if (image_info->interlace == PartitionInterlace)
@@ -655,7 +676,7 @@ static unsigned int WriteRGBImage(const ImageInfo *image_info,Image *image)
             AppendImageFormat("B",image->filename);
             status=OpenBlob(image_info,image,WriteBinaryBlobMode,&image->exception);
             if (status == False)
-              ThrowWriterException(FileOpenError,"UnableToOpenFile",image);
+              ThrowWriterException(FileOpenError,UnableToOpenFile,image);
           }
         if (!MagickMonitor(SaveImageText,200,400,&image->exception))
           break;
@@ -664,7 +685,7 @@ static unsigned int WriteRGBImage(const ImageInfo *image_info,Image *image)
           p=AcquireImagePixels(image,0,y,image->columns,1,&image->exception);
           if (p == (const PixelPacket *) NULL)
             break;
-          (void) PopImagePixels(image,BlueQuantum,pixels);
+          (void) ExportImagePixelArea(image,BlueQuantum,quantum_size,pixels,0,0);
           (void) WriteBlob(image,image->columns,pixels);
         }
         if (LocaleCompare(image_info->magick,"RGBA") == 0)
@@ -678,7 +699,7 @@ static unsigned int WriteRGBImage(const ImageInfo *image_info,Image *image)
                 status=OpenBlob(image_info,image,WriteBinaryBlobMode,
                   &image->exception);
                 if (status == False)
-                  ThrowWriterException(FileOpenError,"UnableToOpenFile",image);
+                  ThrowWriterException(FileOpenError,UnableToOpenFile,image);
               }
             for (y=0; y < (long) image->rows; y++)
             {
@@ -686,12 +707,12 @@ static unsigned int WriteRGBImage(const ImageInfo *image_info,Image *image)
                 &image->exception);
               if (p == (const PixelPacket *) NULL)
                 break;
-              (void) PopImagePixels(image,AlphaQuantum,pixels);
+              (void) ExportImagePixelArea(image,AlphaQuantum,quantum_size,pixels,0,0);
               (void) WriteBlob(image,image->columns,pixels);
             }
           }
         if (image_info->interlace == PartitionInterlace)
-          (void) strncpy(image->filename,image_info->filename,MaxTextExtent-1);
+          (void) strlcpy(image->filename,image_info->filename,MaxTextExtent);
         if (!MagickMonitor(SaveImageText,400,400,&image->exception))
           break;
         break;
@@ -703,7 +724,7 @@ static unsigned int WriteRGBImage(const ImageInfo *image_info,Image *image)
     if (!MagickMonitor(SaveImagesText,scene++,GetImageListLength(image),&image->exception))
       break;
   } while (image_info->adjoin);
-  LiberateMemory((void **) &pixels);
+  MagickFreeMemory(pixels);
   if (image_info->adjoin)
     while (image->previous != (Image *) NULL)
       image=image->previous;
