@@ -18,7 +18,7 @@
 %                            X   X  P      M   M                              %
 %                                                                             %
 %                                                                             %
-%                   Read/Write GraphicsMagick Image Format.                   %
+%                 Read/Write X Windows system Pixmap Format.                  %
 %                                                                             %
 %                                                                             %
 %                              Software Design                                %
@@ -37,7 +37,7 @@
 */
 #include "magick/studio.h"
 #include "magick/blob.h"
-#include "magick/cache.h"
+#include "magick/pixel_cache.h"
 #include "magick/color.h"
 #include "magick/magick.h"
 #include "magick/monitor.h"
@@ -216,12 +216,12 @@ static Image *ReadXPMImage(const ImageInfo *image_info,ExceptionInfo *exception)
   image=AllocateImage(image_info);
   status=OpenBlob(image_info,image,ReadBinaryBlobMode,exception);
   if (status == False)
-    ThrowReaderException(FileOpenError,"UnableToOpenFile",image);
+    ThrowReaderException(FileOpenError,UnableToOpenFile,image);
   /*
     Read XPM file.
   */
   length=MaxTextExtent;
-  xpm_buffer=(char *) AcquireMemory(length);
+  xpm_buffer=MagickAllocateMemory(char *,length);
   p=xpm_buffer;
   if (xpm_buffer != (char *) NULL)
     while (ReadBlobString(image,p) != (char *) NULL)
@@ -235,13 +235,13 @@ static Image *ReadXPMImage(const ImageInfo *image_info,ExceptionInfo *exception)
       if ((p-xpm_buffer+MaxTextExtent+1) < (long) length)
         continue;
       length<<=1;
-      ReacquireMemory((void **) &xpm_buffer,length);
+      MagickReallocMemory(char *,xpm_buffer,length);
       if (xpm_buffer == (char *) NULL)
         break;
       p=xpm_buffer+strlen(xpm_buffer);
     }
   if (xpm_buffer == (char *) NULL)
-    ThrowReaderException(ResourceLimitError,"MemoryAllocationFailed",image);
+    ThrowReaderException(ResourceLimitError,MemoryAllocationFailed,image);
   /*
     Remove comments.
   */
@@ -250,14 +250,14 @@ static Image *ReadXPMImage(const ImageInfo *image_info,ExceptionInfo *exception)
   {
     if (*p != '"')
       continue;
-    count=sscanf(p+1,"%lu %lu %lu %lu",&image->columns,&image->rows,
+    count=sscanf(p+1,"%lu %lu %u %lu",&image->columns,&image->rows,
       &image->colors,&width);
     if (count == 4)
       break;
   }
   if ((count != 4) || (width > 2) || (image->columns == 0) ||
       (image->rows == 0) || (image->colors == 0))
-    ThrowReaderException(CorruptImageError,"NotAXPMImageFile",image)
+    ThrowReaderException(CorruptImageError,ImproperImageHeader,image)
   image->depth=16;
   /*
     Remove unquoted characters.
@@ -277,19 +277,19 @@ static Image *ReadXPMImage(const ImageInfo *image_info,ExceptionInfo *exception)
   }
   xpm_buffer[i]='\0';
   textlist=StringToList(xpm_buffer);
-  LiberateMemory((void **) &xpm_buffer);
+  MagickFreeMemory(xpm_buffer);
   if (textlist == (char **) NULL)
-    ThrowReaderException(ResourceLimitError,"MemoryAllocationFailed",image);
+    ThrowReaderException(ResourceLimitError,MemoryAllocationFailed,image);
   /*
     Initialize image structure.
   */
-  keys=(char **) AcquireMemory(image->colors*sizeof(char *));
+  keys=MagickAllocateMemory(char **,image->colors*sizeof(char *));
   if (!AllocateImageColormap(image,image->colors) || (keys == (char **) NULL))
     {
       for (i=0; textlist[i] != (char *) NULL; i++)
-        LiberateMemory((void **) &textlist[i]);
-      LiberateMemory((void **) &textlist);
-      ThrowReaderException(ResourceLimitError,"MemoryAllocationFailed",image)
+        MagickFreeMemory(textlist[i]);
+      MagickFreeMemory(textlist);
+      ThrowReaderException(ResourceLimitError,MemoryAllocationFailed,image)
     }
   /*
     Read image colormap.
@@ -301,14 +301,14 @@ static Image *ReadXPMImage(const ImageInfo *image_info,ExceptionInfo *exception)
     p=textlist[i++];
     if (p == (char *) NULL)
       break;
-    keys[j]=(char *) AcquireMemory(width+1);
+    keys[j]=MagickAllocateMemory(char *,width+1);
     if (keys[j] == (char *) NULL)
       {
         for (i=0; textlist[i] != (char *) NULL; i++)
-          LiberateMemory((void **) &textlist[i]);
-        LiberateMemory((void **) &textlist);
-        LiberateMemory((void **) &keys);
-        ThrowReaderException(ResourceLimitError,"MemoryAllocationFailed",image)
+          MagickFreeMemory(textlist[i]);
+        MagickFreeMemory(textlist);
+        MagickFreeMemory(keys);
+        ThrowReaderException(ResourceLimitError,MemoryAllocationFailed,image)
       }
     keys[j][width]='\0';
     (void) strncpy(keys[j],p,width);
@@ -321,7 +321,7 @@ static Image *ReadXPMImage(const ImageInfo *image_info,ExceptionInfo *exception)
       {
         while (!isspace((int) (*q)) && (*q != '\0'))
           q++;
-        (void) strncpy(target,q,MaxTextExtent-1);
+        (void) strlcpy(target,q,MaxTextExtent);
         q=ParseColor(target);
         if (q != (char *) NULL)
           *q='\0';
@@ -340,9 +340,9 @@ static Image *ReadXPMImage(const ImageInfo *image_info,ExceptionInfo *exception)
   if (j < (long) image->colors)
     {
       for (i=0; textlist[i] != (char *) NULL; i++)
-        LiberateMemory((void **) &textlist[i]);
-      LiberateMemory((void **) &textlist);
-      ThrowReaderException(CorruptImageError,"CorruptXPMImage",image)
+        MagickFreeMemory(textlist[i]);
+      MagickFreeMemory(textlist);
+      ThrowReaderException(CorruptImageError,CorruptImage,image)
     }
   j=0;
   key[width]='\0';
@@ -358,7 +358,10 @@ static Image *ReadXPMImage(const ImageInfo *image_info,ExceptionInfo *exception)
           break;
         r=SetImagePixels(image,0,y,image->columns,1);
         if (r == (PixelPacket *) NULL)
-          break;
+          {
+            CopyException(exception,&image->exception);
+            break;
+          }
         indexes=GetIndexes(image);
         for (x=0; x < (long) image->columns; x++)
         {
@@ -376,20 +379,23 @@ static Image *ReadXPMImage(const ImageInfo *image_info,ExceptionInfo *exception)
           p+=width;
         }
         if (!SyncImagePixels(image))
-          break;
+          {
+            CopyException(exception,&image->exception);
+            break;
+          }
       }
       if (y < (long) image->rows)
-        ThrowReaderException(CorruptImageError,"NotenoughPixelData",image);
+        ThrowReaderException(CorruptImageError,InsufficientImageDataInFile,image);
     }
   /*
     Free resources.
   */
   for (i=0; i < (long) image->colors; i++)
-    LiberateMemory((void **) &keys[i]);
-  LiberateMemory((void **) &keys);
+    MagickFreeMemory(keys[i]);
+  MagickFreeMemory(keys);
   for (i=0; textlist[i] != (char *) NULL; i++)
-    LiberateMemory((void **) &textlist[i]);
-  LiberateMemory((void **) &textlist);
+    MagickFreeMemory(textlist[i]);
+  MagickFreeMemory(textlist);
   CloseBlob(image);
   return(image);
 }
@@ -434,6 +440,7 @@ ModuleExport void RegisterXPMImage(void)
   entry->decoder=(DecoderHandler) ReadXPMImage;
   entry->encoder=(EncoderHandler) WriteXPMImage;
   entry->adjoin=False;
+  entry->stealth=True;
   entry->description=AcquireString("X Windows system pixmap (color)");
   entry->module=AcquireString("XPM");
   (void) RegisterMagickInfo(entry);
@@ -586,24 +593,39 @@ static unsigned int WritePICONImage(const ImageInfo *image_info,Image *image)
     characters_per_pixel,
     colors;
 
-  /*
-    Open output image file.
-  */
+    ImageCharacteristics
+      characteristics;
+
   assert(image_info != (const ImageInfo *) NULL);
   assert(image_info->signature == MagickSignature);
   assert(image != (Image *) NULL);
   assert(image->signature == MagickSignature);
+  /*
+    Open output image file.
+  */
   status=OpenBlob(image_info,image,WriteBinaryBlobMode,&image->exception);
   if (status == False)
-    ThrowWriterException(FileOpenError,"UnableToOpenFile",image);
-  TransformColorspace(image,RGBColorspace);
+    ThrowWriterException(FileOpenError,UnableToOpenFile,image);
+  /*
+    Ensure that image is in an RGB space.
+  */
+  (void) TransformColorspace(image,RGBColorspace);
+  /*
+    Analyze image to be written.
+  */
+  if (!GetImageCharacteristics(image,&characteristics,
+                               (OptimizeType == image_info->type),
+                               &image->exception))
+    {
+      CloseBlob(image);
+      return MagickFail;
+    }
   SetGeometry(image,&geometry);
   (void) GetMagickGeometry(PiconGeometry,&geometry.x,&geometry.y,
     &geometry.width,&geometry.height);
-  picon=ResizeImage(image,geometry.width,geometry.height,TriangleFilter,1.0,
-    &image->exception);
+  picon=ThumbnailImage(image,geometry.width,geometry.height,&image->exception);
   if ((image_info->type != TrueColorType) &&
-      IsGrayImage(image,&image->exception))
+      (characteristics.grayscale))
     map=BlobToImage(image_info,Graymap,GraymapExtent,&image->exception);
   else
     map=BlobToImage(image_info,Colormap,ColormapExtent,&image->exception);
@@ -645,13 +667,13 @@ static unsigned int WritePICONImage(const ImageInfo *image_info,Image *image)
               break;
           }
         }
-      SetImageType(picon,PaletteType);
+      (void) SetImageType(picon,PaletteType);
     }
   colors=picon->colors;
   if (transparent)
     {
       colors++;
-      ReacquireMemory((void **) &picon->colormap,colors*sizeof(PixelPacket));
+      MagickReallocMemory(PixelPacket *,picon->colormap,colors*sizeof(PixelPacket));
       for (y=0; y < (long) picon->rows; y++)
       {
         q=GetImagePixels(picon,0,y,picon->columns,1);
@@ -733,7 +755,7 @@ static unsigned int WritePICONImage(const ImageInfo *image_info,Image *image)
         symbol[j]=Cixel[k];
       }
       symbol[j]='\0';
-      (void) strncpy(buffer,symbol,MaxTextExtent-1);
+      (void) strlcpy(buffer,symbol,MaxTextExtent);
       (void) WriteBlobString(image,buffer);
     }
     FormatString(buffer,"\"%.1024s\n",
@@ -829,8 +851,8 @@ static unsigned int WriteXPMImage(const ImageInfo *image_info,Image *image)
   assert(image->signature == MagickSignature);
   status=OpenBlob(image_info,image,WriteBinaryBlobMode,&image->exception);
   if (status == False)
-    ThrowWriterException(FileOpenError,"UnableToOpenFile",image);
-  TransformColorspace(image,RGBColorspace);
+    ThrowWriterException(FileOpenError,UnableToOpenFile,image);
+  (void) TransformColorspace(image,RGBColorspace);
   transparent=False;
   if (image->storage_class == PseudoClass)
     {
@@ -865,13 +887,13 @@ static unsigned int WriteXPMImage(const ImageInfo *image_info,Image *image)
               break;
           }
         }
-      SetImageType(image,PaletteType);
+      (void) SetImageType(image,PaletteType);
     }
   colors=image->colors;
   if (transparent)
     {
       colors++;
-      ReacquireMemory((void **) &image->colormap,colors*sizeof(PixelPacket));
+      MagickReallocMemory(PixelPacket *,image->colormap,colors*sizeof(PixelPacket));
       for (y=0; y < (long) image->rows; y++)
       {
         q=GetImagePixels(image,0,y,image->columns,1);
@@ -953,7 +975,7 @@ static unsigned int WriteXPMImage(const ImageInfo *image_info,Image *image)
         symbol[j]=Cixel[k];
       }
       symbol[j]='\0';
-      (void) strncpy(buffer,symbol,MaxTextExtent-1);
+      (void) strlcpy(buffer,symbol,MaxTextExtent);
       (void) WriteBlobString(image,buffer);
     }
     FormatString(buffer,"\"%.1024s\n",
