@@ -36,7 +36,6 @@
   Include declarations.
 */
 #include "magick/studio.h"
-#include "magick/cache.h"
 #include "magick/color.h"
 #include "magick/effect.h"
 #include "magick/enhance.h"
@@ -44,6 +43,7 @@
 #include "magick/gem.h"
 #include "magick/log.h"
 #include "magick/monitor.h"
+#include "magick/pixel_cache.h"
 #include "magick/render.h"
 #include "magick/utility.h"
 
@@ -86,7 +86,7 @@ MagickExport Image *AdaptiveThresholdImage(const Image *image,
   const unsigned long width,const unsigned long height,
   const double offset,ExceptionInfo *exception)
 {
-#define ThresholdImageText  "  Threshold the image...  "
+#define AdaptiveThresholdImageText  "  Threshold the image...  "
 
   DoublePixelPacket
     aggregate,
@@ -119,14 +119,14 @@ MagickExport Image *AdaptiveThresholdImage(const Image *image,
   assert(exception != (ExceptionInfo *) NULL);
   assert(exception->signature == MagickSignature);
   if ((image->columns < width) || (image->rows < height))
-    ThrowImageException(OptionError,"UnableToThresholdImage",
-      "ImageSmallerThanRadius");
+    ThrowImageException3(OptionError,UnableToThresholdImage,
+      ImageSmallerThanRadius);
   threshold_image=CloneImage(image,0,0,True,exception);
   if (threshold_image == (Image *) NULL)
     return((Image *) NULL);
   if (image->is_monochrome)
     return threshold_image;
-  SetImageType(threshold_image,TrueColorType);
+  (void) SetImageType(threshold_image,TrueColorType);
   /*
     Threshold each row of the image.
   */
@@ -167,11 +167,19 @@ MagickExport Image *AdaptiveThresholdImage(const Image *image,
     if (!SyncImagePixels(threshold_image))
       break;
     if (QuantumTick(y,image->rows))
-      if (!MagickMonitor(ThresholdImageText,y,image->rows,exception))
+      if (!MagickMonitor(AdaptiveThresholdImageText,y,image->rows,exception))
         break;
   }
-  threshold_image->is_monochrome=True;
-  threshold_image->is_grayscale=True;
+  if (y < (long) image->rows)
+    {
+      DestroyImage(threshold_image);
+      threshold_image=(Image *) NULL;
+    }
+  else
+    {
+      threshold_image->is_monochrome=True;
+      threshold_image->is_grayscale=True;
+    }
   return(threshold_image);
 }
 
@@ -355,7 +363,7 @@ static void BlurScanline(const double *kernel,const unsigned long width,
               aggregate.blue+=(*p)*q->blue;
               aggregate.opacity+=(*p)*q->opacity;
             }
-          if (((i+width/2-x) >= 0) && ((i+width/2-x) < width))
+          if (((i+(long)(width/2)-x) >= 0) && ((i+width/2-x) < width))
             scale+=kernel[i+width/2-x];
           p++;
           q++;
@@ -457,7 +465,7 @@ static int GetBlurKernel(unsigned long width,const double sigma,double **kernel)
   */
   if (width == 0)
     width=3;
-  *kernel=(double *) AcquireMemory(width*sizeof(double));
+  *kernel=MagickAllocateMemory(double *,width*sizeof(double));
   if (*kernel == (double *) NULL)
     return(0);
   for (i=0; i < (long) width; i++)
@@ -528,37 +536,37 @@ MagickExport Image *BlurImage(const Image *image,const double radius,
       while ((long) (MaxRGB*kernel[0]) > 0)
       {
         if (last_kernel != (double *)NULL)
-          LiberateMemory((void **) &last_kernel);
+          MagickFreeMemory(last_kernel);
         last_kernel=kernel;
         kernel=(double *) NULL;
         width=GetBlurKernel(width+2,sigma,&kernel);
       }
       if (last_kernel != (double *) NULL)
         {
-          LiberateMemory((void **) &kernel);
+          MagickFreeMemory(kernel);
           width-=2;
           kernel=last_kernel;
         }
     }
   if (width < 3)
-    ThrowImageException(OptionError,"UnableToBlurImage",
-      "KernelRadiusIsTooSmall");
+    ThrowImageException3(OptionError,UnableToBlurImage,
+      KernelRadiusIsTooSmall);
   /*
     Allocate blur image.
   */
   blur_image=CloneImage(image,image->columns,image->rows,True,exception);
   if (blur_image == (Image *) NULL)
     {
-      LiberateMemory((void **) &kernel);
+      MagickFreeMemory(kernel);
       return((Image *) NULL);
     }
   blur_image->storage_class=DirectClass;
-  scanline=(PixelPacket *) AcquireMemory(image->rows*sizeof(PixelPacket));
+  scanline=MagickAllocateMemory(PixelPacket *,image->rows*sizeof(PixelPacket));
   if (scanline == (PixelPacket *) NULL)
     {
       DestroyImage(blur_image);
-      ThrowImageException(ResourceLimitError,"MemoryAllocationFailed",
-        "UnableToBlurImage")
+      ThrowImageException(ResourceLimitError,MemoryAllocationFailed,
+        MagickMsg(OptionError,UnableToBlurImage))
     }
   /*
     Blur the image rows.
@@ -600,8 +608,8 @@ MagickExport Image *BlurImage(const Image *image,const double radius,
           break;
       }
   }
-  LiberateMemory((void **) &scanline);
-  LiberateMemory((void **) &kernel);
+  MagickFreeMemory(scanline);
+  MagickFreeMemory(kernel);
   blur_image->is_grayscale=image->is_grayscale;
   return(blur_image);
 }
@@ -634,7 +642,7 @@ MagickExport Image *BlurImage(const Image *image,const double radius,
 MagickExport unsigned int ChannelThresholdImage(Image *image,
   const char *threshold)
 {
-#define ThresholdImageText  "  Threshold the image...  "
+#define ChannelThresholdImageText  "  Threshold the image...  "
 
   double
     red_threshold,
@@ -691,7 +699,7 @@ MagickExport unsigned int ChannelThresholdImage(Image *image,
   blue_threshold_quantum=RoundSignedToQuantum(blue_threshold);
   opacity_threshold_quantum=RoundSignedToQuantum(opacity_threshold);
 
-  SetImageType(image,TrueColorType);
+  (void) SetImageType(image,TrueColorType);
 
   for (y=0; y < (long) image->rows; y++)
   {
@@ -724,7 +732,7 @@ MagickExport unsigned int ChannelThresholdImage(Image *image,
     if (!SyncImagePixels(image))
       break;
     if (QuantumTick(y,image->rows))
-      if (!MagickMonitor(ThresholdImageText,y,image->rows,&image->exception))
+      if (!MagickMonitor(ChannelThresholdImageText,y,image->rows,&image->exception))
         break;
   }
 
@@ -812,13 +820,13 @@ MagickExport Image *DespeckleImage(const Image *image,ExceptionInfo *exception)
     Allocate image buffers.
   */
   length=(image->columns+2)*(image->rows+2)*sizeof(Quantum);
-  pixels=(Quantum *) AcquireMemory(length);
-  buffer=(Quantum *) AcquireMemory(length);
+  pixels=MagickAllocateMemory(Quantum *,length);
+  buffer=MagickAllocateMemory(Quantum *,length);
   if ((buffer == (Quantum *) NULL) || (pixels == (Quantum *) NULL))
     {
       DestroyImage(despeckle_image);
-      ThrowImageException(ResourceLimitError,"MemoryAllocationFailed",
-        "UnableToDespeckleImage")
+      ThrowImageException3(ResourceLimitError,MemoryAllocationFailed,
+        UnableToDespeckleImage)
     }
   /*
     Reduce speckle in the image.
@@ -904,8 +912,8 @@ MagickExport Image *DespeckleImage(const Image *image,ExceptionInfo *exception)
   /*
     Free memory.
   */
-  LiberateMemory((void **) &buffer);
-  LiberateMemory((void **) &pixels);
+  MagickFreeMemory(buffer);
+  MagickFreeMemory(pixels);
   despeckle_image->is_grayscale=image->is_grayscale;
   return(despeckle_image);
 }
@@ -960,17 +968,17 @@ MagickExport Image *EdgeImage(const Image *image,const double radius,
   assert(exception->signature == MagickSignature);
   width=GetOptimalKernelWidth(radius,0.5);
   if (((long) image->columns < width) || ((long) image->rows < width))
-    ThrowImageException(OptionError,"UnableToEdgeImage",
-      "ImageIsSmallerThanRadius");
-  kernel=(double *) AcquireMemory(width*width*sizeof(double));
+    ThrowImageException3(OptionError,UnableToEdgeImage,
+      ImageSmallerThanRadius);
+  kernel=MagickAllocateMemory(double *,width*width*sizeof(double));
   if (kernel == (double *) NULL)
-    ThrowImageException(ResourceLimitError,"MemoryAllocationFailed",
-      "UnableToEdgeImage");
+    ThrowImageException3(ResourceLimitError,MemoryAllocationFailed,
+      UnableToEdgeImage);
   for (i=0; i < (width*width); i++)
     kernel[i]=(-1.0);
   kernel[i/2]=width*width-1.0;
   edge_image=ConvolveImage(image,width,kernel,exception);
-  LiberateMemory((void **) &kernel);
+  MagickFreeMemory(kernel);
   edge_image->is_grayscale=image->is_grayscale;
   return(edge_image);
 }
@@ -1032,10 +1040,10 @@ MagickExport Image *EmbossImage(const Image *image,const double radius,
   assert(exception != (ExceptionInfo *) NULL);
   assert(exception->signature == MagickSignature);
   width=GetOptimalKernelWidth(radius,0.5);
-  kernel=(double *) AcquireMemory(width*width*sizeof(double));
+  kernel=MagickAllocateMemory(double *,width*width*sizeof(double));
   if (kernel == (double *) NULL)
-    ThrowImageException(ResourceLimitError,"MemoryAllocationFailed",
-      "UnableToEmbossImage");
+    ThrowImageException3(ResourceLimitError,MemoryAllocationFailed,
+      UnableToEmbossImage);
   i=0;
   j=width/2;
   for (v=(-width/2); v <= (width/2); v++)
@@ -1054,7 +1062,7 @@ MagickExport Image *EmbossImage(const Image *image,const double radius,
   emboss_image=ConvolveImage(image,width,kernel,exception);
   if (emboss_image != (Image *) NULL)
     (void) EqualizeImage(emboss_image);
-  LiberateMemory((void **) &kernel);
+  MagickFreeMemory(kernel);
   emboss_image->is_grayscale=image->is_grayscale;
   return(emboss_image);
 }
@@ -1271,12 +1279,12 @@ MagickExport Image *GaussianBlurImage(const Image *image,const double radius,
   assert(exception->signature == MagickSignature);
   width=GetOptimalKernelWidth2D(radius,sigma);
   if (((long) image->columns < width) || ((long) image->rows < width))
-    ThrowImageException(OptionError,"UnableToBlurImage",
-      "ImageIsSmallerThanRadius");
-  kernel=(double *) AcquireMemory(width*width*sizeof(double));
+    ThrowImageException3(OptionError,UnableToBlurImage,
+      ImageSmallerThanRadius);
+  kernel=MagickAllocateMemory(double *,width*width*sizeof(double));
   if (kernel == (double *) NULL)
-    ThrowImageException(ResourceLimitError,"MemoryAllocationFailed",
-      "UnableToBlurImage");
+    ThrowImageException(ResourceLimitError,MemoryAllocationFailed,
+      MagickMsg(OptionError,UnableToBlurImage));
   i=0;
   for (v=(-width/2); v <= (width/2); v++)
   {
@@ -1288,7 +1296,7 @@ MagickExport Image *GaussianBlurImage(const Image *image,const double radius,
     }
   }
   blur_image=ConvolveImage(image,width,kernel,exception);
-  LiberateMemory((void **) &kernel);
+  MagickFreeMemory(kernel);
   blur_image->is_grayscale=image->is_grayscale;
   return(blur_image);
 }
@@ -1342,7 +1350,7 @@ typedef struct _MedianSkipList
     level;
 
   MedianListNode
-    nodes[65537];
+    *nodes;
 } MedianSkipList;
 
 typedef struct _MedianPixelList
@@ -1458,11 +1466,34 @@ static PixelPacket GetMedianList(MedianPixelList *pixel_list)
   return(pixel);
 }
 
-static void InitializeMedianList(MedianPixelList *pixel_list,long width)
+static int InitializeMedianList(MedianPixelList *pixel_list,long width)
 {
+  register int i;
   pixel_list->center=width*width/2;
   pixel_list->signature=MagickSignature;
   (void) memset((void *) pixel_list->lists,0,4*sizeof(MedianSkipList));
+  for(i=0;i<4;i++)
+    {
+      pixel_list->lists[i].nodes =
+        MagickAllocateMemory(MedianListNode *,65537 * sizeof(MedianListNode));
+      if(!pixel_list->lists[i].nodes)
+	{
+	  assert(0);
+	  return (False);
+	}
+      (void) memset(pixel_list->lists[i].nodes,0,65537U*sizeof(MedianListNode));
+    }
+  return (True);
+}
+
+static void CleanMedianList(MedianPixelList *pixel_list)
+{
+  register int i;
+  for(i=0;i<4;i++)
+    {
+      if(pixel_list->lists[i].nodes)
+        MagickFreeMemory(pixel_list->lists[i].nodes);
+    }
 }
 
 static inline void InsertMedianList(MedianPixelList *pixel_list,
@@ -1570,8 +1601,8 @@ MagickExport Image *MedianFilterImage(const Image *image,const double radius,
   assert(exception->signature == MagickSignature);
   width=GetOptimalKernelWidth(radius,0.5);
   if (((long) image->columns < width) || ((long) image->rows < width))
-    ThrowImageException(OptionError,"UnableToFilterImage",
-      "ImageSmallerThanKernelRadius");
+    ThrowImageException3(OptionError,UnableToFilterImage,
+      ImageSmallerThanRadius);
   median_image=CloneImage(image,image->columns,image->rows,True,exception);
   if (median_image == (Image *) NULL)
     return((Image *) NULL);
@@ -1579,17 +1610,23 @@ MagickExport Image *MedianFilterImage(const Image *image,const double radius,
   /*
     Allocate skip-lists.
   */
-  skiplist=(MedianPixelList *) AcquireMemory(sizeof(MedianPixelList));
+  skiplist=MagickAllocateMemory(MedianPixelList *,sizeof(MedianPixelList));
   if (skiplist == (MedianPixelList *) NULL)
     {
       DestroyImage(median_image);
-      ThrowImageException(ResourceLimitError,"MemoryAllocationFailed",
-        "UnableToMedianFilterImage")
+      ThrowImageException3(ResourceLimitError,MemoryAllocationFailed,
+        UnableToMedianFilterImage)
     }
   /*
     Median filter each image row.
   */
-  InitializeMedianList(skiplist,width);
+  if(!InitializeMedianList(skiplist,width))
+  {
+    DestroyImage(median_image);
+    ThrowImageException3(ResourceLimitError,MemoryAllocationFailed,
+	  UnableToMedianFilterImage)
+  }
+
   for (y=0; y < (long) median_image->rows; y++)
   {
     p=AcquireImagePixels(image,-width/2,y-width/2,image->columns+width,width,
@@ -1616,7 +1653,8 @@ MagickExport Image *MedianFilterImage(const Image *image,const double radius,
       if (!MagickMonitor(MedianFilterImageText,y,median_image->rows,exception))
         break;
   }
-  LiberateMemory((void **) &skiplist);
+  CleanMedianList(skiplist);
+  MagickFreeMemory(skiplist);
   median_image->is_grayscale=image->is_grayscale;
   return(median_image);
 }
@@ -1681,7 +1719,7 @@ static int GetMotionBlurKernel(int width,const double sigma,double **kernel)
   */
   if (width <= 0)
     width=3;
-  *kernel=(double *) AcquireMemory(width*sizeof(double));
+  *kernel=MagickAllocateMemory(double *,width*sizeof(double));
   if (*kernel == (double *) NULL)
     return(0);
   for (i=0; i < width; i++)
@@ -1750,33 +1788,33 @@ MagickExport Image *MotionBlurImage(const Image *image,const double radius,
       while ((MaxRGB*kernel[width-1]) > 0.0)
       {
         if (last_kernel != (double *)NULL)
-          LiberateMemory((void **) &last_kernel);
+          MagickFreeMemory(last_kernel);
         last_kernel=kernel;
         kernel=(double *) NULL;
         width=GetMotionBlurKernel(width+2,sigma,&kernel);
       }
       if (last_kernel != (double *) NULL)
         {
-          LiberateMemory((void **) &kernel);
+          MagickFreeMemory(kernel);
           width-=2;
           kernel=last_kernel;
         }
     }
   if (width < 3)
-    ThrowImageException(OptionError,"UnableToBlurimage",
-      "KernelRadiusIsTooSmall");
-  offsets=(PointInfo *) AcquireMemory(width*sizeof(PointInfo));
+    ThrowImageException3(OptionError,UnableToBlurImage,
+      KernelRadiusIsTooSmall);
+  offsets=MagickAllocateMemory(PointInfo *,width*sizeof(PointInfo));
   if (offsets == (PointInfo *) NULL)
-    ThrowImageException(ResourceLimitError,"MemoryAllocationFailed",
-      "UnableToMotionBlurImage");
+    ThrowImageException3(ResourceLimitError,MemoryAllocationFailed,
+      UnableToMotionBlurImage);
   /*
     Allocate blur image.
   */
   blur_image=CloneImage(image,image->columns,image->rows,True,exception);
   if (blur_image == (Image *) NULL)
     {
-      LiberateMemory((void **) &kernel);
-      LiberateMemory((void **) &offsets);
+      MagickFreeMemory(kernel);
+      MagickFreeMemory(offsets);
       return((Image *) NULL);
     }
   blur_image->storage_class=DirectClass;
@@ -1821,8 +1859,8 @@ MagickExport Image *MotionBlurImage(const Image *image,const double radius,
       if (!MagickMonitor(BlurImageText,y,image->rows,exception))
         break;
   }
-  LiberateMemory((void **) &kernel);
-  LiberateMemory((void **) &offsets);
+  MagickFreeMemory(kernel);
+  MagickFreeMemory(offsets);
   blur_image->is_grayscale=image->is_grayscale;
   return(blur_image);
 }
@@ -1939,8 +1977,8 @@ MagickExport unsigned int RandomChannelThresholdImage(Image *image,const char
   if (LocaleCompare(channel,"all") == 0 ||
       LocaleCompare(channel,"intensity") == 0)
     if (!AllocateImageColormap(image,2))
-      ThrowBinaryException(ResourceLimitError,"MemoryAllocationFailed",
-        "UnableToThresholdImage");
+      ThrowBinaryException3(ResourceLimitError,MemoryAllocationFailed,
+        UnableToThresholdImage);
 
   for (y=0; y < (long) image->rows; y++)
   {
@@ -2042,8 +2080,8 @@ MagickExport unsigned int RandomChannelThresholdImage(Image *image,const char
       {
         /* To Do: red, green, blue, cyan, magenta, yellow, black */
         if (LocaleCompare(channel,"intensity") != 0)
-          ThrowBinaryException(OptionError, "UnableToThresholdimage",
-              "UnrecognizedChannelType");
+          ThrowBinaryException3(OptionError, UnableToThresholdImage,
+              UnrecognizedChannelType);
       }
 
     if (!SyncImagePixels(image))
@@ -2180,8 +2218,8 @@ MagickExport Image *ReduceNoiseImage(const Image *image,const double radius,
   assert(exception->signature == MagickSignature);
   width=GetOptimalKernelWidth(radius,0.5);
   if (((long) image->columns < width) || ((long) image->rows < width))
-    ThrowImageException(OptionError,"UnableToFilterImage",
-      "ImageSmallerThanKernelRadius");
+    ThrowImageException3(OptionError,UnableToFilterImage,
+      ImageSmallerThanRadius);
   noise_image=CloneImage(image,image->columns,image->rows,True,exception);
   if (noise_image == (Image *) NULL)
     return((Image *) NULL);
@@ -2189,17 +2227,23 @@ MagickExport Image *ReduceNoiseImage(const Image *image,const double radius,
   /*
     Allocate skip-lists.
   */
-  skiplist=(MedianPixelList *) AcquireMemory(sizeof(MedianPixelList));
+  skiplist=MagickAllocateMemory(MedianPixelList *,sizeof(MedianPixelList));
   if (skiplist == (MedianPixelList *) NULL)
     {
       DestroyImage(noise_image);
-      ThrowImageException(ResourceLimitError,"MemoryAllocationFailed",
-        "UnableToNoiseFilterImage")
+      ThrowImageException3(ResourceLimitError,MemoryAllocationFailed,
+        UnableToNoiseFilterImage)
     }
   /*
     Median filter each image row.
   */
-  InitializeMedianList(skiplist,width);
+  if(!InitializeMedianList(skiplist,width))
+  {
+    DestroyImage(noise_image);
+    ThrowImageException3(ResourceLimitError,MemoryAllocationFailed,
+	  UnableToMedianFilterImage)
+  }
+
   for (y=0; y < (long) noise_image->rows; y++)
   {
     p=AcquireImagePixels(image,-width/2,y-width/2,image->columns+width,width,
@@ -2228,7 +2272,8 @@ MagickExport Image *ReduceNoiseImage(const Image *image,const double radius,
       if (!MagickMonitor(ReduceNoiseImageText,y,noise_image->rows,exception))
         break;
   }
-  LiberateMemory((void **) &skiplist);
+  CleanMedianList(skiplist);
+  MagickFreeMemory(skiplist);
   noise_image->is_grayscale=image->is_grayscale;
   return(noise_image);
 }
@@ -2446,12 +2491,12 @@ MagickExport Image *SharpenImage(const Image *image,const double radius,
   assert(exception->signature == MagickSignature);
   width=GetOptimalKernelWidth(radius,sigma);
   if (((long) image->columns < width) || ((long) image->rows < width))
-    ThrowImageException(OptionError,"UnableToSharpenImage",
-      "ImageIsSmallerThanRadius");
-  kernel=(double *) AcquireMemory(width*width*sizeof(double));
+    ThrowImageException3(OptionError,UnableToSharpenImage,
+      ImageSmallerThanRadius);
+  kernel=MagickAllocateMemory(double *,width*width*sizeof(double));
   if (kernel == (double *) NULL)
-    ThrowImageException(ResourceLimitError,"MemoryAllocationFailed",
-      "UnableToSharpenImage");
+    ThrowImageException3(ResourceLimitError,MemoryAllocationFailed,
+      UnableToSharpenImage);
   i=0;
   normalize=0.0;
   for (v=(-width/2); v <= (width/2); v++)
@@ -2466,7 +2511,7 @@ MagickExport Image *SharpenImage(const Image *image,const double radius,
   }
   kernel[i/2]=(-2.0)*normalize;
   sharp_image=ConvolveImage(image,width,kernel,exception);
-  LiberateMemory((void **) &kernel);
+  MagickFreeMemory(kernel);
   sharp_image->is_grayscale=image->is_grayscale;
   return(sharp_image);
 }
@@ -2511,12 +2556,12 @@ MagickExport Image *SpreadImage(const Image *image,const unsigned int radius,
   int
     offsets_index;
 
-  long
+  int
     x_distance,
     y,
     y_distance;
 
-  register long
+  register int
     x;
 
   long
@@ -2546,14 +2591,13 @@ MagickExport Image *SpreadImage(const Image *image,const unsigned int radius,
   /*
     Initialize random offsets cache
   */
-  offsets=(long *)AcquireMemory(OFFSETS_ENTRIES*sizeof(long));
+  offsets=MagickAllocateMemory(long *,OFFSETS_ENTRIES*sizeof(long));
   if (offsets == (long *) NULL)
     {
-      ThrowException(exception,ResourceLimitError,
-                  "MemoryAllocationFailed",NULL);
+      ThrowException(exception,ResourceLimitError,MemoryAllocationFailed,NULL);
       return (Image *) NULL;
     }
-  for (x=0; x < (OFFSETS_ENTRIES-1); x++)
+  for (x=0; x < OFFSETS_ENTRIES; x++)
     offsets[x]=((((2*(double) radius+1)*rand())/RAND_MAX)-((long)radius));
 
   /*
@@ -2569,7 +2613,7 @@ MagickExport Image *SpreadImage(const Image *image,const unsigned int radius,
     if (q == (PixelPacket *) NULL)
       break;
 
-    if (radius > y)
+    if (radius > (unsigned int) y)
       y_min=0;
     else
       y_min=y-radius;
@@ -2608,7 +2652,7 @@ MagickExport Image *SpreadImage(const Image *image,const unsigned int radius,
       if (!MagickMonitor(SpreadImageText,y,image->rows,exception))
         break;
   }
-  LiberateMemory((void **)&offsets);
+  MagickFreeMemory(offsets);
   spread_image->is_grayscale=image->is_grayscale;
   spread_image->is_monochrome=image->is_monochrome;
   return(spread_image);
@@ -2625,12 +2669,14 @@ MagickExport Image *SpreadImage(const Image *image,const unsigned int radius,
 %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 %
 %  ThresholdImage() changes the value of individual pixels based on
-%  the intensity of each pixel compared to threshold.  The result is a
-%  high-contrast, two color image.
+%  the intensity of each pixel compared to a specified threshold.  Values
+%  greater than the threshold are set to the maximum quantum value, and
+%  values equal to or below the threshold are set to the minimum quantum
+%  value.  The result is a high-contrast, two color image.
 %
 %  The format of the ThresholdImage method is:
 %
-%      unsigned int ThresholdImage(Image *image,const double threshold)
+%      MagickPassFail ThresholdImage(Image *image,const double threshold)
 %
 %  A description of each parameter follows:
 %
@@ -2640,96 +2686,98 @@ MagickExport Image *SpreadImage(const Image *image,const unsigned int radius,
 %
 %
 */
-MagickExport unsigned int ThresholdImage(Image *image,const double threshold)
+MagickExport MagickPassFail ThresholdImage(Image *image,const double threshold)
 {
-#define ThresholdImageText  "  Threshold the image...  "
+#define ThresholdImageText  "Threshold the image...  "
 
   register IndexPacket
     index;
 
-  long
+  unsigned long
     y;
 
   register IndexPacket
     *indexes;
 
-  register unsigned long
+  Quantum
+    intensity,
     quantum_threshold;
 
-  register long
+  register unsigned long
     x;
 
   register PixelPacket
     *q;
 
+  MagickBool
+    grayscale,
+    initialize_indexes,
+    modified;
+
   /*
     Threshold image.
   */
   assert(image != (Image *) NULL);
   assert(image->signature == MagickSignature);
 
+  initialize_indexes=(image->storage_class == PseudoClass ? MagickFalse : MagickTrue);
+  grayscale=image->is_grayscale;
+  quantum_threshold=RoundDoubleToQuantum(threshold);
+
+  /*
+    Check if image is already in desired state.
+  */
+  if ((quantum_threshold != MaxRGB) && (image->storage_class == PseudoClass)
+      && (image->colors == 2))
+    {
+      if (IsBlackPixel(image->colormap[0]) && (IsWhitePixel(image->colormap[1])))
+        {
+          image->is_monochrome=MagickTrue;
+          image->is_grayscale=MagickTrue;
+          return MagickPass;
+        }
+    }
+
   if (!AllocateImageColormap(image,2))
-    ThrowBinaryException(ResourceLimitError,"MemoryAllocationFailed",
-      "UnableToThresholdImage");
+    ThrowBinaryException3(ResourceLimitError,MemoryAllocationFailed,
+                          UnableToThresholdImage);
 
-  quantum_threshold=RoundSignedToQuantum(threshold);
-
-  for (y=0; y < (long) image->rows; y++)
-  {
-    q=GetImagePixels(image,0,y,image->columns,1);
-    if (q == (PixelPacket *) NULL)
-      break;
-    indexes=GetIndexes(image);
-    if (quantum_threshold == MaxRGB)
-      {
-        /* All pixels thresholded to black */
-        for (x=(long) image->columns; x > 0; x--)
-          {
-            *indexes++=q->red=q->green=q->blue=0;
-            q++;
-          }
-      }
-    else if (quantum_threshold == 0)
-      {
-        /* All pixels thresholded to white */
-        for (x=(long) image->columns; x > 0; x--)
-          {
-            *indexes++=1;
-            q->red=q->green=q->blue=MaxRGB;
-            q++;
-          }
-      }
-    else if (image->is_grayscale)
-      {
-        for (x=(long) image->columns; x > 0; x--)
-          {
-            /* Image is grayscale so q->red contains intensity */
-            index=q->red <= quantum_threshold ? 0 : 1;
-            *indexes++=index;
-            q->red=q->green=q->blue=image->colormap[index].red;
-            q++;
-          }
-      }
-    else
-      {
-        for (x=(long) image->columns; x > 0; x--)
-          {
-            /* Compute itensity value of RGB pixel */
-            index=PixelIntensityToQuantum(q) <= quantum_threshold ? 0 : 1;
-            *indexes++=index;
-            q->red=q->green=q->blue=image->colormap[index].red;
-            q++;
-          }
-      }
-    if (!SyncImagePixels(image))
-      break;
-    if (QuantumTick(y,image->rows))
-      if (!MagickMonitor(ThresholdImageText,y,image->rows,&image->exception))
+  for (y=0; y < image->rows; y++)
+    {
+      q=GetImagePixels(image,0,y,image->columns,1);
+      if (q == (PixelPacket *) NULL)
         break;
-  }
-  image->is_monochrome=True;
-  image->is_grayscale=True;
-  return(True);
+      indexes=GetIndexes(image);
+      modified=MagickFalse;
+
+      for (x=0; x < image->columns; x++)
+        {
+          if (grayscale)
+            intensity=q[x].red;
+          else
+            intensity=PixelIntensityToQuantum(&q[x]);
+          index=(intensity <= quantum_threshold ? 0U : 1U);
+          if ((initialize_indexes) || (index != indexes[x]))
+            {
+              modified=MagickTrue;
+              indexes[x]=index;
+            }
+          if (NotColorMatch(&image->colormap[index],&q[x]))
+            {
+              modified=MagickTrue;
+              q[x].red=q[x].green=q[x].blue=image->colormap[index].red;
+            }
+        }
+      if (modified)
+        if (!SyncImagePixels(image))
+          break;
+      if (QuantumTick(y,image->rows))
+        if (!MagickMonitor(ThresholdImageText,y,image->rows,&image->exception))
+          break;
+    }
+  image->is_monochrome=MagickTrue;
+  image->is_grayscale=MagickTrue;
+  return(MagickPass);
 }
 
 /*
