@@ -37,35 +37,38 @@
 */
 #include "magick/studio.h"
 #include "magick/utility.h"
-#if defined(HasPTHREADS)
+#if defined(HAVE_PTHREAD)
 #include <pthread.h>
 #endif
-#if defined(WIN32)
+#if defined(MSWINDOWS)
 #include <windows.h>
+#define USE_SPINLOCKS
+#define SPINLOCK_DELAY_MILLI_SECS 10
 #endif
+#include "magick/semaphore.h"
 
 /*
   Struct declaractions.
 */
-struct SemaphoreInfo
+struct _SemaphoreInfo
 {
-#if defined(HasPTHREADS)
+#if defined(HAVE_PTHREAD)
   pthread_mutex_t
     mutex;		/* POSIX thread mutex */
 
   pthread_t
-    thread_id;		/* ID of thread which holds the lock */
+    owner_thread_id;    /* ID of thread which holds the lock */
 #endif
-#if defined(WIN32)
+#if defined(MSWINDOWS)
   CRITICAL_SECTION
     mutex;		/* Windows critical section */
 
   DWORD
-    thread_id;		/* ID of thread which holds the lock */
+    owner_thread_id;    /* ID of thread which holds the lock */
 #endif
 
   unsigned int
-    locked;		/* True if semaphore is locked */
+    lock_depth;         /* Number of times lock is held by same thread */
 
   unsigned long
     signature;		/* Used to validate structure */
@@ -74,17 +77,48 @@ struct SemaphoreInfo
 /*
   Static declaractions.
 */
-#if defined(HasPTHREADS)
+#if defined(HAVE_PTHREAD)
 static pthread_mutex_t
   semaphore_mutex = PTHREAD_MUTEX_INITIALIZER;
 #endif
 
-#if defined(WIN32)
+#if defined(MSWINDOWS)
+#if !defined(USE_SPINLOCKS)
 static CRITICAL_SECTION
   semaphore_mutex;
 
 static unsigned int
   active_semaphore = False;
+#else
+static LONG
+  semaphore_mutex = 0;
+/* Wait for spin lock */
+static void spinlock_wait (LONG volatile *sl)
+{
+  /*
+    InterlockedCompareExchange performs an atomic comparison of the
+    specified 32-bit values and exchanges them, based on the outcome of
+    the comparison. Requires Windows XP, Windows 2000 Professional,
+    Windows NT Workstation 4.0, Windows Me, or Windows 98.
+   */
+
+  while (InterlockedCompareExchange (sl, 1L, 0L) != 0)
+  {
+    /* slight delay - just in case OS does not giveup CPU */
+    Sleep (SPINLOCK_DELAY_MILLI_SECS);
+  }
+}
+/* Release spin lock */
+static void spinlock_release (LONG volatile *sl)
+{
+  /*
+    InterlockedExchange atomically exchanges a pair of 32-bit
+    values. Requires Windows XP, Windows 2000 Professional, Windows NT
+    Workstation 3.5 and later, Windows Me, Windows 98, or Windows 95.
+  */
+  InterlockedExchange (sl, 0L);
+}
+#endif
 #endif
 
 /*
@@ -98,7 +132,8 @@ static unsigned int
 %                                                                             %
 %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 %
-%  AcquireSemaphoreInfo() acquires a semaphore.
+%  AcquireSemaphoreInfo() locks a semaphore, initializing it first if
+%  necessary.
 %
 %  The format of the AcquireSemaphoreInfo method is:
 %
@@ -113,22 +148,30 @@ static unsigned int
 MagickExport void AcquireSemaphoreInfo(SemaphoreInfo **semaphore_info)
 {
   assert(semaphore_info != (SemaphoreInfo **) NULL);
-#if defined(HasPTHREADS)
+#if defined(HAVE_PTHREAD)
   (void) pthread_mutex_lock(&semaphore_mutex);
 #endif
-#if defined(WIN32)
+#if defined(MSWINDOWS)
+#if !defined(USE_SPINLOCKS)
   if (!active_semaphore)
     InitializeCriticalSection(&semaphore_mutex);
   active_semaphore=True;
   EnterCriticalSection(&semaphore_mutex);
+#else
+  spinlock_wait(&semaphore_mutex);
+#endif
 #endif
   if (*semaphore_info == (SemaphoreInfo *) NULL)
     *semaphore_info=AllocateSemaphoreInfo();
-#if defined(HasPTHREADS)
+#if defined(HAVE_PTHREAD)
   (void) pthread_mutex_unlock(&semaphore_mutex);
 #endif
-#if defined(WIN32)
+#if defined(MSWINDOWS)
+#if !defined(USE_SPINLOCKS)
   LeaveCriticalSection(&semaphore_mutex);
+#else
+  spinlock_release(&semaphore_mutex);
+#endif
 #endif
   (void) LockSemaphoreInfo(*semaphore_info);
 }
@@ -165,32 +208,50 @@ MagickExport SemaphoreInfo *AllocateSemaphoreInfo(void)
   /*
     Allocate semaphore.
   */
-  semaphore_info=(SemaphoreInfo *) AcquireMemory(sizeof(SemaphoreInfo));
+  semaphore_info=MagickAllocateMemory(SemaphoreInfo *,sizeof(SemaphoreInfo));
   if (semaphore_info == (SemaphoreInfo *) NULL)
-    MagickFatalError(ResourceLimitFatalError,"MemoryAllocationFailed",
-      "UnableToAllocateSemaphoreInfo");
-  memset(semaphore_info,0,sizeof(SemaphoreInfo));
+    MagickFatalError3(ResourceLimitFatalError,MemoryAllocationFailed,
+      UnableToAllocateSemaphoreInfo);
+  (void) memset(semaphore_info,0,sizeof(SemaphoreInfo));
   /*
     Initialize the semaphore.
   */
-#if defined(HasPTHREADS)
+#if defined(HAVE_PTHREAD)
   {
     int
       status;
 
-    status=pthread_mutex_init(&semaphore_info->mutex,
-      (const pthread_mutexattr_t *) NULL);
-    if (status != 0)
+    pthread_mutexattr_t 
+      mutexattr;
+
+    if ((status=pthread_mutexattr_init(&mutexattr)) != 0)
       {
-        LiberateMemory((void **) &semaphore_info);
+        MagickFreeMemory(semaphore_info);
+        errno=status;
         return((SemaphoreInfo *) NULL);
       }
+
+#if defined(PTHREAD_MUTEX_RECURSIVE)
+    (void) pthread_mutexattr_settype(&mutexattr, PTHREAD_MUTEX_RECURSIVE);
+#endif /* PTHREAD_MUTEX_RECURSIVE */
+
+    if ((status=pthread_mutex_init(&semaphore_info->mutex,&mutexattr)) != 0)
+      {
+        MagickFreeMemory(semaphore_info);
+        errno=status;
+        return((SemaphoreInfo *) NULL);
+      }
+
+    (void) pthread_mutexattr_destroy(&mutexattr);
+
+    semaphore_info->owner_thread_id=0;
   }
 #endif
-#if defined(WIN32)
+#if defined(MSWINDOWS)
   InitializeCriticalSection(&semaphore_info->mutex);
 #endif
-  semaphore_info->locked=False;
+  semaphore_info->lock_depth=0;
+  
   semaphore_info->signature=MagickSignature;
   return(semaphore_info);
 }
@@ -216,14 +277,16 @@ MagickExport SemaphoreInfo *AllocateSemaphoreInfo(void)
 */
 MagickExport void DestroySemaphore(void)
 {
-#if defined(HasPTHREADS)
+#if defined(HAVE_PTHREAD)
   (void) pthread_mutex_destroy(&semaphore_mutex);
 #endif
-#if defined(WIN32)
+#if defined(MSWINDOWS)
+#if !defined(USE_SPINLOCKS)
   if (active_semaphore)
     DeleteCriticalSection(&semaphore_mutex);
   active_semaphore=False;
 #endif
+#endif
 }
 
 /*
@@ -255,28 +318,35 @@ MagickExport void DestroySemaphoreInfo(SemaphoreInfo **semaphore_info)
   if (*semaphore_info == (SemaphoreInfo *) NULL)
     return;
   assert((*semaphore_info)->signature == MagickSignature);
-  (void) UnlockSemaphoreInfo(*semaphore_info);
-#if defined(HasPTHREADS)
+#if defined(HAVE_PTHREAD)
   (void) pthread_mutex_lock(&semaphore_mutex);
 #endif
-#if defined(WIN32)
+#if defined(MSWINDOWS)
+#if !defined(USE_SPINLOCKS)
   if (!active_semaphore)
     InitializeCriticalSection(&semaphore_mutex);
   active_semaphore=True;
   EnterCriticalSection(&semaphore_mutex);
+#else
+  spinlock_wait(&semaphore_mutex);
 #endif
-#if defined(HasPTHREADS)
+#endif
+#if defined(HAVE_PTHREAD)
   (void) pthread_mutex_destroy(&(*semaphore_info)->mutex);
 #endif
-#if defined(WIN32)
+#if defined(MSWINDOWS)
   DeleteCriticalSection(&(*semaphore_info)->mutex);
 #endif
-  LiberateMemory((void **) &(*semaphore_info));
-#if defined(HasPTHREADS)
+  MagickFreeMemory((*semaphore_info));
+#if defined(HAVE_PTHREAD)
   (void) pthread_mutex_unlock(&semaphore_mutex);
 #endif
-#if defined(WIN32)
+#if defined(MSWINDOWS)
+#if !defined(USE_SPINLOCKS)
   LeaveCriticalSection(&semaphore_mutex);
+#else
+  spinlock_release(&semaphore_mutex);
+#endif
 #endif
 }
 
@@ -301,15 +371,17 @@ MagickExport void DestroySemaphoreInfo(SemaphoreInfo **semaphore_info)
 */
 MagickExport void InitializeSemaphore(void)
 {
-#if defined(HasPTHREADS)
+#if defined(HAVE_PTHREAD)
   (void) pthread_mutex_init(&semaphore_mutex,
     (const pthread_mutexattr_t *) NULL);
 #endif
-#if defined(WIN32)
+#if defined(MSWINDOWS)
+#if !defined(USE_SPINLOCKS)
   if (!active_semaphore)
     InitializeCriticalSection(&semaphore_mutex);
   active_semaphore=True;
 #endif
+#endif
 }
 
 /*
@@ -323,7 +395,7 @@ MagickExport void InitializeSemaphore(void)
 %                                                                             %
 %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 %
-%  Method LiberateSemaphoreInfo liberates a semaphore.
+%  Method LiberateSemaphoreInfo unlocks the semaphore if it has been allocated.
 %
 %  The format of the LiberateSemaphoreInfo method is:
 %
@@ -370,23 +442,44 @@ MagickExport void LiberateSemaphoreInfo(SemaphoreInfo **semaphore_info)
 %
 %
 */
-MagickExport unsigned int LockSemaphoreInfo(SemaphoreInfo *semaphore_info)
+MagickExport MagickPassFail LockSemaphoreInfo(SemaphoreInfo *semaphore_info)
 {
   assert(semaphore_info != (SemaphoreInfo *) NULL);
   assert(semaphore_info->signature == MagickSignature);
-#if defined(HasPTHREADS)
-  if (pthread_mutex_lock(&semaphore_info->mutex))
-    return(False);
-  /* Record the thread ID of the locking thread */
-  semaphore_info->thread_id=pthread_self();
+
+#if defined(HAVE_PTHREAD)
+  {
+    int
+      status;
+
+    pthread_t
+      self;
+
+    self=pthread_self();
+
+    if ((semaphore_info->lock_depth > 0) &&
+        (semaphore_info->owner_thread_id == self))
+      {
+        fprintf(stderr,"Warning: recursive semaphore lock detected!\n");
+        fflush(stderr);
+      }
+    
+    if ((status = pthread_mutex_lock(&semaphore_info->mutex)) != 0)
+      {
+        errno=status;
+        return(MagickFail);
+      }
+    /* Record the thread ID of the locking thread */
+    semaphore_info->owner_thread_id=self;
+  }
 #endif
-#if defined(WIN32)
+#if defined(MSWINDOWS)
   EnterCriticalSection(&semaphore_info->mutex);
   /* Record the thread ID of the locking thread */
-  semaphore_info->thread_id=GetCurrentThreadId();
+  semaphore_info->owner_thread_id=GetCurrentThreadId();
 #endif
-  semaphore_info->locked=True;
-  return(True);
+  semaphore_info->lock_depth++;
+  return(MagickPass);
 }
 
 /*
@@ -415,23 +508,32 @@ MagickExport unsigned int LockSemaphoreInfo(SemaphoreInfo *semaphore_info)
 %
 %
 */
-MagickExport unsigned int UnlockSemaphoreInfo(SemaphoreInfo *semaphore_info)
+MagickExport MagickPassFail UnlockSemaphoreInfo(SemaphoreInfo *semaphore_info)
 {
   assert(semaphore_info != (SemaphoreInfo *) NULL);
   assert(semaphore_info->signature == MagickSignature);
-  if (semaphore_info->locked != True)
-    return (False);
-#if defined(HasPTHREADS)
+
+  if (semaphore_info->lock_depth == 0)
+    {
+      fprintf(stderr,
+              "Warning: unlock on unlocked semaphore (p=%p)!\n",
+              semaphore_info);
+      fflush(stderr);
+      return (MagickFail);
+    }
+
+  semaphore_info->lock_depth--;
+  
+#if defined(HAVE_PTHREAD)
   /* Enforce that unlocking thread is the same as the locking thread */
-  assert(pthread_equal(semaphore_info->thread_id,pthread_self()));
+  assert(pthread_equal(semaphore_info->owner_thread_id,pthread_self()));
   if (pthread_mutex_unlock(&semaphore_info->mutex))
-    return(False);
+    return(MagickFail);
 #endif
-#if defined(WIN32)
+#if defined(MSWINDOWS)
   /* Enforce that unlocking thread is the same as the locking thread */
-  assert(GetCurrentThreadId() == semaphore_info->thread_id);
+  assert(GetCurrentThreadId() == semaphore_info->owner_thread_id);
   LeaveCriticalSection(&semaphore_info->mutex);
 #endif
-  semaphore_info->locked=False;
-  return(True);
+  return(MagickPass);
 }
